//! Little Librarian - Semantic search for Calibre libraries.

use std::io;

use clap::Parser;
use little_librarian::{
    api,
    cli::{Cli, Commands},
    query,
    scanner::{self},
    storage,
    text_encoder::{self, TextEncoder},
    tokenize::{self, Tokenizer},
};
use snafu::{ResultExt, Snafu};
use tracing_subscriber::{layer::SubscriberExt, util::SubscriberInitExt};

/// Top-level application errors.
#[derive(Debug, Snafu)]
pub enum Error {
    #[snafu(display("Failed to read db-url from stdin."))]
    ReadStdin { source: io::Error },
    #[snafu(display("Failed to read db-url from file '{file}'."))]
    ReadFile { source: io::Error, file: String },
    #[snafu(display("Failed to load tokenizer."))]
    TokenizerLoading { source: tokenize::FromFileError },
    #[snafu(display("Failed to load embedding model."))]
    ModelLoading {
        source: text_encoder::NewFromFileError,
    },
    #[snafu(display("Failed to load reranker model."))]
    RerankerLoading {
        source: text_encoder::NewFromFileError,
    },
    #[snafu(display("Failed to connect to database."))]
    DbConnection { source: storage::ConnectionError },
    #[snafu(display("Failed to process query."))]
    Ask { source: query::AskError },
    #[snafu(display("Failed to serve HTTP API."))]
    Serve { source: api::ServeError },
}

#[tokio::main]
#[snafu::report]
async fn main() -> Result<(), Error> {
    tracing_subscriber::registry()
        .with(
            tracing_subscriber::EnvFilter::try_from_default_env().unwrap_or_else(|_| {
                format!(
                    "{}=debug,tower_http=debug,axum::rejection=trace",
                    env!("CARGO_CRATE_NAME")
                )
                .into()
            }),
        )
        .with(tracing_subscriber::fmt::layer())
        .init();

    let cli = Cli::parse();
    let db_url = if cli.db_url == "-" {
        io::read_to_string(io::stdin()).context(ReadStdinSnafu)?
    } else {
        std::fs::read_to_string(&cli.db_url).context(ReadFileSnafu { file: cli.db_url })?
    };

    let tokenizer = Tokenizer::from_file(&cli.tokenizer_path).context(TokenizerLoadingSnafu)?;
    let embedder = TextEncoder::from_file(&cli.model_path).context(ModelLoadingSnafu)?;
    let db = storage::Postgres::connect(&db_url, 10)
        .await
        .context(DbConnectionSnafu)?;

    match &cli.command {
        Commands::Ask(config) => {
            let reranker =
                TextEncoder::from_file(&config.reranking_path).context(RerankerLoadingSnafu)?;
            query::ask(
                &config.query,
                &db,
                &tokenizer,
                &embedder,
                &reranker,
                cli.chunk_size,
                config.limit,
            )
            .await
            .context(AskSnafu)?;
        }
        Commands::Scan(config) => {
            scanner::scan(
                &db,
                &config.library_path,
                &tokenizer,
                &embedder,
                cli.chunk_size,
            )
            .await;
        }
        Commands::Serve(config) => {
            let reranker =
                TextEncoder::from_file(&config.reranking_path).context(RerankerLoadingSnafu)?;
            api::serve(
                &config.address,
                db,
                tokenizer,
                embedder,
                reranker,
                cli.chunk_size,
            )
            .await
            .context(ServeSnafu)?;
        }
    }

    Ok(())
}
