//! Book scanning and processing for embedding generation.

use std::{fs::File, io, path::PathBuf};

use snafu::{ResultExt, Snafu};
use tracing::{error, info};

use crate::{
    calibre::{Book, BookIterator},
    extractors::{self, extractor::ExtractionError},
    hash,
    storage::{self, Postgres},
    text_encoder::{self, TextEncoder},
    tokenize::{self, Tokenizer},
};

/// Scan a Calibre library and process all books for embedding generation.
pub async fn scan(
    db: &Postgres,
    library_path: &PathBuf,
    tokenizer: &Tokenizer,
    text_encoder: &TextEncoder,
    chunk_size: usize,
) {
    let books = BookIterator::new(library_path);
    for book_result in books {
        match book_result {
            Ok(book) => {
                if let Err(e) = process_book(&book, db, tokenizer, text_encoder, chunk_size).await {
                    error!("Failed to process book {}: {e}", book.id);
                }
            }
            Err(e) => {
                error!("error fetching calibre book: {e}");
            }
        }
    }
}

/// Errors that occur during book processing.
#[derive(Debug, Snafu)]
pub enum ProcessBookError {
    #[snafu(display("Failed to hash book."))]
    BookHash { source: io::Error },
    #[snafu(display("Failed to check in db if document needs update."))]
    UpdateCheck { source: storage::CheckUpdateError },
    #[snafu(display("Failed to open book."))]
    Open { source: io::Error },
    #[snafu(display("Failed to extract document content."))]
    Extraction { source: ExtractionError },
    #[snafu(display("Failed to tokenize document content."))]
    Encode { source: tokenize::EncodeError },
    #[snafu(display("Failed to create embeddings from encodings."))]
    Embedding { source: text_encoder::EmbedError },
    #[snafu(display("Failed to look up document versions."))]
    DocumentVersion { source: storage::VersionError },
    #[snafu(display("Failed to store embeddings in database."))]
    SaveEmbeddings { source: storage::EmbeddingError },
    #[snafu(display("Failed to update document hash in database."))]
    UpdateHash { source: storage::UpdateHashError },
    #[snafu(display("Failed to delete old document versions in database."))]
    RemoveOldVersions {
        source: storage::RemoveOldVersionsError,
    },
}

/// Process a single book: extract text, generate embeddings, and store in database.
async fn process_book(
    book: &Book,
    db: &Postgres,
    tokenizer: &Tokenizer,
    embedder: &TextEncoder,
    chunk_size: usize,
) -> Result<(), ProcessBookError> {
    let book_hash = hash::file_sha256(&book.path).context(BookHashSnafu)?;

    if db
        .book_needs_update(book.id, book_hash)
        .await
        .context(UpdateCheckSnafu)?
    {
        info!("processing {:?}", book.path);

        let input = File::open(book.path.clone()).context(OpenSnafu)?;
        if let Some(text) =
            extractors::extractor::extract(&input, &book.extension).context(ExtractionSnafu)?
        {
            let encodings = tokenizer.encode(&text, chunk_size).context(EncodeSnafu)?;
            let embeddings = embedder.embed_multiple(encodings).context(EmbeddingSnafu)?;
            let version = db
                .embedding_version(book.id)
                .await
                .context(DocumentVersionSnafu)?
                + 1;

            db.add_multiple_embeddings(book.id, version, embeddings)
                .await
                .context(SaveEmbeddingsSnafu)?;
        }

        db.update_hash(book.id, book_hash)
            .await
            .context(UpdateHashSnafu)?;
        db.remove_old_versions(book.id)
            .await
            .context(RemoveOldVersionsSnafu)?;
    }

    Ok(())
}
