//! Text tokenization and chunking for embedding models.

use std::path::Path;

use snafu::{ResultExt, Snafu};

/// Errors that occur when loading a tokenizer from file.
#[derive(Debug, Snafu)]
pub enum FromFileError {
    #[snafu(display("Failed to read tokenizer file."))]
    File { source: std::io::Error },
    #[snafu(display("Failed to load tokenizer."))]
    New { source: FromBytesError },
}

/// Error when creating tokenizer from bytes.
#[derive(Debug, Snafu)]
#[snafu(display("Failed to read tokenizer from file."))]
pub struct FromBytesError {
    source: tokenizers::Error,
}

/// Errors that occur during text encoding.
#[derive(Debug, Snafu)]
pub enum EncodeError {
    #[snafu(display("Failed to encode."))]
    Encode { source: tokenizers::Error },
}

/// Wrapper around a tokenizer that handles text chunking.
#[derive(Debug, Clone)]
pub struct Tokenizer {
    /// The underlying tokenizer implementation.
    inner: tokenizers::Tokenizer,
}

/// A tokenized encoding paired with its original text.
pub type Encoding = (tokenizers::Encoding, String);

impl Tokenizer {
    /// Create a tokenizer from a file path.
    pub fn from_file<P: AsRef<Path>>(path: P) -> Result<Self, FromFileError> {
        let contents = std::fs::read(path).context(FileSnafu)?;
        Self::new(&contents).context(NewSnafu)
    }

    /// Create a tokenizer from configuration bytes.
    pub fn new<P: AsRef<[u8]>>(config: P) -> Result<Self, FromBytesError> {
        let inner = tokenizers::Tokenizer::from_bytes(config).context(FromBytesSnafu)?;
        Ok(Self { inner })
    }

    /// Normalize whitespace and remove extra spaces from text.
    fn clean_text(text: &str) -> String {
        text.chars()
            .map(|c| match c {
                '\n' | '\r' | '\t' => ' ',
                c if c.is_whitespace() => ' ',
                c => c,
            })
            .collect::<String>()
            .split_whitespace()
            .collect::<Vec<&str>>()
            .join(" ")
    }

    /// Split text into word-based chunks and encode each chunk.
    pub fn encode(&self, text: &str, chunk_size: usize) -> Result<Vec<Encoding>, EncodeError> {
        let text = Self::clean_text(text);
        let words: Vec<&str> = text.split_whitespace().collect();

        words
            .chunks(chunk_size)
            .map(|word_chunk| word_chunk.join(" "))
            .map(|chunk| {
                let tokenizer_impl = &self.inner;
                tokenizer_impl
                    .encode(chunk.clone(), false)
                    .context(EncodeSnafu)
                    .map(|e| (e, chunk))
            })
            .collect::<Result<Vec<Encoding>, EncodeError>>()
    }
}
